from django.template import Template, Context, TemplateSyntaxError
from django.template.loader import get_template
from django.test import SimpleTestCase
from django.utils.safestring import SafeString


class TestSandwichTag(SimpleTestCase):
    def setUp(self):
        self.default_title = "Why a Penguin should be the next president"
        self.default_fixings = "<p>They dress well, and make everyone happy.</p>"
        self.sw_template_name = "bread.html"

    def _sandwich_with_fixings(
        self,
        bread: str | Template,
        title: str | None = None,
        fixings: str | None = None,
        with_title: bool = True,
        bread_as_var: bool = False,
    ) -> Template:
        bread_template = bread if bread_as_var and bread else f"'{bread}'" if bread else ""
        title = f" title='{self.default_title if title is None else title}'" if with_title else ""
        return Template(
            "{% load sandwich %}"
            "{% sandwich " + bread_template + title + " %}"
            f"{self.default_fixings if fixings is None else fixings}"
            "{% endsandwich %}"
        )

    def test_template_not_provided(self):
        self.assertRaises(
            TemplateSyntaxError,
            self._sandwich_with_fixings,
            bread="",
        )

    def test_with_template_object(self):
        """the template passed to a sandwich tag may be a string or Template instance"""
        rendered = self._sandwich_with_fixings("bread_template", bread_as_var=True, title="", fixings="").render(
            Context({"bread_template": get_template(self.sw_template_name).template})
        )
        self.assertHTMLEqual(rendered, "<h1>default title text</h1><div></div>")

    def test_with_template_name(self):
        with_sandwich = self._sandwich_with_fixings(self.sw_template_name).render(Context({}))
        manual = get_template(self.sw_template_name).template.render(  # not treated as a sandwich
            Context(dict(title=self.default_title, sandwich_fixings=SafeString(self.default_fixings)))
        )
        self.assertHTMLEqual(with_sandwich, manual)

    def test_bad_bread_template_type(self):
        """try passing a non-string/non-Template instance as bread"""
        bad_bread_var = "bad_bread"
        context = Context()
        context[bad_bread_var] = {}  # not a string or Template instance
        template = self._sandwich_with_fixings(bad_bread_var, bread_as_var=True)
        self.assertRaises(TemplateSyntaxError, template.render, context)

    def test_sandwich_context_isolation(self):  # alternate name: test_sandwich_context_behaviour
        """
        Ensure that the...
        - parent template only receives context from kwargs passed to the tag
        - fixings template only receives global context (implicit)
        """
        context = Context()
        context["title"] = "custom title text"
        context["body"] = "custom body text"
        rendered = self._sandwich_with_fixings(
            bread=self.sw_template_name,
            fixings="{{ body|default:'default body text' }}",
            with_title=False,
        ).render(context)
        self.assertNotIn("custom title text", rendered)
        self.assertIn("custom body text", rendered)

    def test_resolved_kwargs_not_cached(self):
        """
        Ensure that kwargs are re-resolved on each call to the tag.

        If resolved kwargs are stored in the instance, then they will be cached and
        will not be resolved again.

        That means that if a sandwich tag is in a for-loop block, each iteration
        would use the same values as the first iteration, where the kwargs were resolved.

        That's not what we want.
        """
        context_dict = {
            "titles": [
                "title1",
                "title2",
                "title3",
            ]
        }
        # expected = "\n".join([f"<h1>{d['my_title']}</h1><div></div>" for d in context_dict['kwargs']])
        expected = "<h1>title1</h1><div></div>" "<h1>title2</h1><div></div>" "<h1>title3</h1><div></div>"
        rendered = Template(
            "{% load sandwich %}"
            "{% for my_title in titles %}"
            "{% sandwich 'bread.html' title=my_title %}" 
            "{% endsandwich %}"
            "{% endfor %}"
        ).render(Context(context_dict))
        self.assertHTMLEqual(rendered, expected)

    def test_sandwich_renders_children(self):
        """
        No imposters allowed. Real sandwiches only.
        """
        rendered = self._sandwich_with_fixings(
            bread=self.sw_template_name,
            fixings="{{ body|default:'default body text' }}",
            title="",
        ).render(Context({}))
        self.assertHTMLEqual(rendered, "<h1>default title text</h1><div>default body text</div>")


from functools import wraps
from inspect import unwrap, getfullargspec

from django.template import Template
from django.utils.itercompat import is_iterable
from django.template.library import Library, parse_bits, TagHelperNode
from django.template.backends.django import Template as BackendTemplate


DEFAULT_CHILD_VAR_NAME = 'sandwich_fixings'
DEFAULT_TAKES_CONTEXT = False
DEFAULT_BREAD_USES_GLOBAL_CONTEXT = False
DEFAULT_CHILDREN_USE_GLOBAL_CONTEXT = True

def get_compile_func(
        func,
        template: str,
        name: str = None,
        takes_context=DEFAULT_TAKES_CONTEXT,
        bread_uses_global_context=DEFAULT_BREAD_USES_GLOBAL_CONTEXT,
        children_use_global_context=DEFAULT_CHILDREN_USE_GLOBAL_CONTEXT,
        child_var_name=DEFAULT_CHILD_VAR_NAME,
):
    function_name = name or func.__name__
    (
        params,
        varargs,
        varkw,
        defaults,
        kwonly,
        kwonly_defaults,
        _,
    ) = getfullargspec(unwrap(func))
    if takes_context:
        assert params[0] == 'context', (
            'First param of custom sandwich tag must be `context` if `takes_context` is `True`'
        )

    def compile_func(parser, token):
        open_sw_tag, *bits = token.split_contents()
        close_sw_tag = "end" + open_sw_tag
        child_nodelist = parser.parse(parse_until=(close_sw_tag,))
        parser.delete_first_token()

        token_args, token_kwargs = parse_bits(
            parser=parser,
            bits=bits,
            params=params,
            varargs=varargs,
            varkw=varkw,
            defaults=defaults,
            kwonly=kwonly,
            kwonly_defaults=kwonly_defaults,
            takes_context=takes_context,
            name=open_sw_tag,
        )
        return SandwichTagNode(
            func=func,
            args=token_args,
            kwargs=token_kwargs,
            filename=template,
            child_nodelist=child_nodelist,
            takes_context=takes_context,
            child_var_name=child_var_name,
            bread_uses_global_context=bread_uses_global_context,
            children_use_global_context=children_use_global_context,
        )

    compile_func.__name__ = function_name
    return compile_func


def register_sandwich_tag(
        template: str,
        registry: Library = None,
        name: str = None,
        func=None,
        takes_context=DEFAULT_TAKES_CONTEXT,
        bread_uses_global_context=DEFAULT_BREAD_USES_GLOBAL_CONTEXT,
        children_use_global_context=DEFAULT_CHILDREN_USE_GLOBAL_CONTEXT,
        child_var_name=DEFAULT_CHILD_VAR_NAME,
):
    def dec(func):

        @wraps(func)
        def compile_func(parser, token):
            return get_compile_func(
                func=func,
                template=template,
                name=name,
                takes_context=takes_context,
                child_var_name=child_var_name,
                bread_uses_global_context=bread_uses_global_context,
                children_use_global_context=children_use_global_context,
            )(parser, token)

        registry.tag(compile_func.__name__, compile_func)
        return func

    if func is None:
        return dec
    elif callable(func):
        return dec(func)
    else:
        raise ValueError("Invalid arguments provided to register_sandwich_tag")


def add_sandwich_tag_dec(registry: Library):
    def sandwich_tag(
            template: str,
            name: str = None,
            func=None,
            takes_context=DEFAULT_TAKES_CONTEXT,
            child_var_name=DEFAULT_CHILD_VAR_NAME,
            bread_uses_global_context=DEFAULT_BREAD_USES_GLOBAL_CONTEXT,
            children_use_global_context=DEFAULT_CHILDREN_USE_GLOBAL_CONTEXT,
    ):
        return register_sandwich_tag(
            template=template,
            registry=registry,
            name=name,
            func=func,
            takes_context=takes_context,
            bread_uses_global_context=bread_uses_global_context,
            children_use_global_context=children_use_global_context,
            child_var_name=child_var_name,
        )

    registry.sandwich_tag = sandwich_tag
    return registry


def get_and_set_template_from_spec(spec, context, node):
    t = context.render_context.get(node)
    if t is None:
        if isinstance(spec, (Template, BackendTemplate)):
            t = spec
        elif isinstance(getattr(spec, "template", None), Template):
            t = spec.template
        elif not isinstance(spec, str) and is_iterable(spec):
            t = context.template.engine.select_template(spec)
        else:
            t = context.template.engine.get_template(spec)
        context.render_context[node] = t
    return t


class SandwichTagNode(TagHelperNode):
    def __init__(
            self,
            func,
            args,
            kwargs,
            filename,
            child_nodelist,
            takes_context=DEFAULT_TAKES_CONTEXT,
            bread_uses_global_context=DEFAULT_BREAD_USES_GLOBAL_CONTEXT,
            children_use_global_context=DEFAULT_CHILDREN_USE_GLOBAL_CONTEXT,
            child_var_name=DEFAULT_CHILD_VAR_NAME,
    ):
        super().__init__(func, takes_context, args, kwargs)
        self.filename = filename
        self.child_nodelist = child_nodelist
        self.child_var_name = child_var_name
        self.bread_uses_global_context = bread_uses_global_context
        self.children_use_global_context = children_use_global_context

    def render(self, context):
        bread_template = get_and_set_template_from_spec(self.filename, context, self)

        resolved_args, resolved_kwargs = self.get_resolved_arguments(context)
        _dict = self.func(*resolved_args, **resolved_kwargs)
        if isinstance(_dict, dict) and self.child_var_name in _dict:
            child_dict = _dict[self.child_var_name]
        else:
            child_dict = {}

        if self.children_use_global_context:
            with context.push(child_dict):
                rendered_children = self.child_nodelist.render(context)
        else:
            rendered_children = self.child_nodelist.render(context.new(child_dict))

        if self.bread_uses_global_context:
            with context.push(_dict):
                context[self.child_var_name] = rendered_children
                return bread_template.render(context)

        bread_context = context.new(_dict)
        bread_context[self.child_var_name] = rendered_children
        return bread_template.render(bread_context)



from typing import TypeAlias

from django import template
from django.template import Template, Context
from django.template.base import FilterExpression, Parser, Token
from django.template.library import Library, parse_bits


register = Library()


TemplateSpec: TypeAlias = str | Template


def resolve_template_spec(template_spec: TemplateSpec, context: Context) -> Template:
    match template_spec:
        case str():
            if '{{' in template_spec or '{%' in template_spec:
                return Template(template_spec, engine=context.template.engine)
            return context.template.engine.get_template(template_spec)
        case Template():
            return template_spec
        case _:
            raise template.TemplateSyntaxError(
                f"template param must be a string or a Template object, got {template_spec} instead."
            )


@register.tag(name="sandwich")
def do_sandwich(parser: Parser, token: Token) -> template.Node:  # do_{tag} follows djangos convention
    """
    Processes the `sandwich` template tag, which wraps child content in a parent template.

    The `sandwich` tag allows dynamic rendering of a parent template (specified as a string or
    Template object) with a placeholder (`{{ sandwich_fixings }}`) where the child content
    renders. Keyword arguments passed to the tag define the parent template's context, while
    the child template accesses the global context implicitly.

    Args:
        parser (Parser): The template parser instance.
        token (Token): The parsed token containing the tag name and its arguments.

    Returns:
        SandwichNode: A template node that renders the parent template with the child content.

    Raises:
        TemplateSyntaxError: If the `template` argument is missing, invalid, or specified
        both positionally and as a keyword.
    """

    # first bit is tag name, no need to parse it
    open_sw_tag, *bits = token.split_contents()
    close_sw_tag = "end" + open_sw_tag
    child_nodelist = parser.parse(parse_until=(close_sw_tag,))
    parser.delete_first_token()
    takes_context = False
    if any(context_arg := [bits.pop(i) for i, arg in enumerate(bits) if arg == 'context']):
        assert len(context_arg) == 1, "`context` special arg appeared more than once."
        takes_context = True
    token_args, token_kwargs = parse_bits(
        parser=parser,
        bits=bits,
        params=("template",),  # Note that `params` must be an iterable (not `None`)
        varargs=["context"],  # use None if no varargs (DONT USE `False` since it's checked as `if varargs is None`)
        varkw='kwargs',  # accept token_kwargs that aren't in params (like `def func(**token_kwargs)`)
        defaults=None,
        kwonly=(),  # kwonly must be an iterable, we don't have any
        kwonly_defaults=None,
        takes_context=False,  # if true, first elt of params must be `"context"`
        # Note that `sandwich` does take context implicitly in order for the "fixings" to access it, but currently,
        # the parent template has no access to it
        name=open_sw_tag,
    )
    # `parent_fe` is the FilterExpression instance that "knows" the name of the parent template
    # Note that a string with no variables or filters is stored in a FilterExpression instance by django

    if token_args:  # the only other token_arg allowed is the `template`
        if "template" in token_kwargs:
            raise template.TemplateSyntaxError(
                f"{open_sw_tag} tag received template as both positional and keyword argument. "
                "Please remove one of them."
            )
        bread_spec = token_args[0]
    else:
        bread_spec = token_kwargs.pop("template")
    return SandwichNode(child_nodelist, bread_spec, token_kwargs, takes_context)


class SandwichNode(template.Node):
    """passes rendered tag contents to the bread template as context named sandwich_fixings"""

    child_var_name = "sandwich_fixings"

    def __init__(self, child_nodelist: template.NodeList, filter_expression: FilterExpression, token_kwargs, takes_context: bool):
        self.child_nodelist = child_nodelist
        self.filter_expression = filter_expression
        self.token_kwargs = token_kwargs
        self.takes_context = takes_context

    def _get_bread_template(self, context: Context) -> Template:
        template_spec = self.filter_expression.resolve(context)
        return resolve_template_spec(template_spec, context)

    def _resolve_kwargs(self, context) -> dict:
        return {k: getattr(v, "resolve", lambda c: v)(context) for k, v in self.token_kwargs.items()}

    def render_bread(self, template, context: Context, rendered_fixings, overrides) -> str:
        with context.push(overrides):
            context[self.child_var_name] = rendered_fixings
            return template.render(context)

    def render(self, context: Context):
        _dict = self._resolve_kwargs(context)
        template = self._get_bread_template(context)  # use global context to resolve the name of the template
        rendered_children = self.child_nodelist.render(context)
        if self.takes_context:
            print("Using global context")
            return self.render_bread(template, context, rendered_children, _dict)
        print("Using local context")
        return self.render_bread(template, context.new(), rendered_children, _dict)
